@@ -1,31 +1,71 @@
 '''
-    Estimates the CRE model and computes posterior using Trace
-    Approximation.
+Defines class CREEstimator, which uses a trace approximation to estimate the CRE model.
 '''
 # import logging
 # from pathlib import Path
 import pyamg
 import numpy as np
 import pandas as pd
-from bipartitepandas import update_dict, logger_init
+from bipartitepandas.util import ParamsDict, logger_init
 from scipy.sparse import csc_matrix, coo_matrix, diags, linalg, eye
 import time
 # import pyreadr
 # import os
 # from multiprocessing import Pool, TimeoutError
 # from timeit import default_timer as timer
-
 # import argparse
 import json
 # import itertools
+from tqdm.auto import tqdm, trange
 
-# try to use tqdm
-try:
-    from tqdm import tqdm, trange
-except ImportError:
-    trange = range
+# NOTE: multiprocessing isn't compatible with lambda functions
+def _gteq1(a):
+    return a >= 1
 
-def pipe_qcov(df, e1, e2):
+# Define default parameter dictionary
+_cre_params_default = ParamsDict({
+    'ncore': (1, 'type_constrained', (int, _gteq1),
+        '''
+            (default=1) Number of cores to use.
+        ''', '>= 1'),
+    'posterior': (False, 'type', bool,
+        '''
+            (default=False) If True, compute posterior variance.
+        ''', None),
+    'wo_btw': (False, 'type', bool,
+        '''
+            (default=False) If True, sets between variation to 0, pure RE.
+        ''', None),
+    'ndraw_trace': (5, 'type_constrained', (int, _gteq1),
+        '''
+            (default=5) Number of draws to use in trace approximations.
+        ''', '>= 1'),
+    'outputfile': (None, 'type_none', str,
+        '''
+            (default=None) Outputfile where results will be saved in json format. If None, results will not be saved.
+        ''', None)
+    # 'ndp': (50, 'type_constrained', (int, _gteq1), # FIXME not used
+    #     '''
+    #         (default=50) Number of draws to use in approximation of leverages.
+    #     ''', '>= 1')
+})
+
+def cre_params(update_dict=None):
+    '''
+    Dictionary of default cre_params. Run tw.cre_params().describe_all() for descriptions of all valid parameters.
+
+    Arguments:
+        update_dict (dict or None): user parameter values; None is equivalent to {}
+
+    Returns:
+        (ParamsDict) dictionary of cre_params
+    '''
+    new_dict = _cre_params_default.copy()
+    if update_dict is not None:
+        new_dict.update(update_dict)
+    return new_dict
+
+def _pipe_qcov(df, e1, e2):
     v1 = df.eval(e1)
     v2 = df.eval(e2)
     return(np.cov(v1, v2)[0][1])
@@ -34,7 +74,7 @@ def pipe_qcov(df, e1, e2):
 #     rows = itertools.product(*data_dict.values())
 #     return pd.DataFrame.from_records(rows, columns=data_dict.keys())
 
-def pd_to_np(df, colr, colc, colv, nr, nc):
+def _pd_to_np(df, colr, colc, colv, nr, nc):
     row_index = df[colr].to_numpy()
     col_index = df[colc].to_numpy()
     values = df[colv].to_numpy()
@@ -42,18 +82,18 @@ def pd_to_np(df, colr, colc, colv, nr, nc):
 
     for i in range(nr):
         for j in range(nc):
-            I = (row_index == i + 1) & (col_index == j + 1)
+            I = (row_index == i) & (col_index == j)
             if I.sum() > 0:
                 A[i, j] = values[I][0]
 
     return(A)
-    # pd_to_np(df, 'i', 'j', 'v', 3, 3)
+    # _pd_to_np(df, 'i', 'j', 'v', 3, 3)
 
 class CREEstimator:
     '''
     Uses multigrid and partialing out to solve two way Fixed Effect model.
     '''
-    def __init__(self, data, params):
+    def __init__(self, data, params=None):
         '''
         Arguments:
             data (Pandas DataFrame): cross-section labor data. Data contains the following columns:
@@ -79,68 +119,60 @@ class CREEstimator:
                 m (0 if stayer, 1 if mover)
 
                 cs (0 if not in cross section, 1 if in cross section)
-            params (dict): dictionary of parameters for CRE estimation
-
-                Dictionary parameters:
-
-                    ncore (int, default=1): number of cores to use
-
-                    ndraw_tr (int, default=5): number of draws to use in approximation for traces
-
-                    ndp (int, default=50): number of draw to use in approximation for leverages
-
-                    out (str, default='res_cre.json'): outputfile where results are saved
-
-                    posterior (bool, default=False): if True, compute posterior variance
-
-                    wo_btw (bool, default=False): if True, sets between variation to 0, pure RE
+            cre_params (ParamsDict or None): dictionary of parameters for CRE estimation. Run tw.cre_params().describe_all() for descriptions of all valid parameters. None is equivalent to tw.cre_params().
         '''
         # Start logger
         logger_init(self)
         # self.logger.info('initializing CREEstimator object')
 
+        if params is None:
+            params = cre_params()
+
         self.adata = data
 
-        # Define default parameter dictionaries
-        default_params = {
-            'ncore': 1, # Number of cores to use
-            'ndraw_tr': 5, # Number of draws to use in approximation for traces
-            'ndp': 50, # Number of draw to use in approximation for leverages
-            'out': 'res_cre.json', # Outputfile where results are saved
-            'posterior': False, # If True, compute posterior variance
-            'wo_btw': False # If True, sets between variation to 0, pure RE
-        }
-
-        self.params = update_dict(default_params, params)
+        self.params = params
         self.res = {} # Results dictionary
         self.summary = {} # Summary results dictionary
 
-        # Save some commonly used parameters as attributes
-        self.ncore = self.params['ncore'] # Number of cores to use
-        self.ndraw_trace = self.params['ndraw_tr'] # Number of draws to use in approximation for traces
-        self.wo_btw = self.params['wo_btw'] # If True, sets between variation to 0, pure RE
+        ## Save some commonly used parameters as attributes
+        # Number of cores to use
+        self.ncore = params['ncore']
+        # Number of draws to use in trace approximations
+        self.ndraw_trace = params['ndraw_trace']
+        # If True, sets between variation to 0, pure RE
+        self.wo_btw = params['wo_btw']
 
-        # Store some parameters in results dictionary
+        ## Store some parameters in results dictionary
         self.res['cores'] = self.ncore
         self.res['ndt'] = self.ndraw_trace
 
         # self.logger.info('CREEstimator object initialized')
 
-    def fit(self):
+    def fit(self, rng=None):
         '''
         Run CRE solver.
+
+        Arguments:
+            rng (np.random.Generator or None): NumPy random number generator; None is equivalent to np.random.default_rng(None)
         '''
+        if rng is None:
+            rng = np.random.default_rng(None)
+
+        self.logger.info('----- STARTING CRE ESTIMATION -----')
         self.start_time = time.time()
 
         # Begin cleaning and analysis
         self.__prep_vars() # Prepare data
 
         # Generate stayers and movers, and set indices so they don't overlap
-        jdata = self.adata[(self.adata['m'] == 1) & (self.adata['cs'] == 1)].reset_index(drop=True)
+        jdata = self.adata[(self.adata['m'] > 0) & (self.adata['cs'] == 1)].reset_index(drop=True)
         self.mn = len(jdata) # Number of observations from movers # FIXME I renamed from nm to mn, since nm makes it seem like it's the number of movers, while mn gives movers, where n is the total number of observations
-        sdata = self.adata[(self.adata['m'] == 0) & (self.adata['cs'] == 1)].set_index(np.arange(self.ns) + self.mn) # FIXME was np.arange(self.ns) + 1 + self.mn
+        ########################################
+        ##### NOTE: if you code crashes on the next line, it is likely because your data is not collapsed #####
+        ########################################
+        sdata = self.adata[(self.adata['m'] == 0) & (self.adata['cs'] == 1)].set_index(np.arange(self.ns) + self.mn) # FIXME could be np.arange(len(sdata)) + self.mn for non-collapsed data, since stayers can have multiple observations - but this has issues at the moment because consecutive observations at the same firm for movers should be marked as stayers
         sdata, jdata = self.__estimate_between_cluster(sdata, jdata)
-        self.__estimate_within_cluster(sdata, jdata)
+        self._estimate_within_cluster(sdata, jdata)
         self.__estimate_within_parameters()
 
         cdata = pd.concat([sdata[['y1', 'psi1_tmp', 'mx', 'j1']], jdata[['y1', 'psi1_tmp', 'mx', 'j1']]], axis=0)
@@ -154,7 +186,7 @@ class CREEstimator:
             self.__prep_posterior_var(jdata, cdata)
 
             self.logger.info('computing posterior variance')
-            self.__compute_posterior_var()
+            self.__compute_posterior_var(rng)
             self.res['var_posterior'] = self.posterior_var
 
         end_time = time.time()
@@ -164,7 +196,7 @@ class CREEstimator:
 
         self.__save_res() # Save results to json
 
-        self.logger.info('------ DONE -------')
+        self.logger.info('----- DONE WITH CRE ESTIMATION -----')
 
     def __prep_vars(self):
         '''
@@ -196,12 +228,11 @@ class CREEstimator:
         self.nf = max(self.adata['j1'].max(), self.adata['j2'].max()) + 1 # Number of firms
         self.nw = self.adata['i'].max() + 1 # Number of workers
         self.nc = max(self.adata['g1'].max(), self.adata['g2'].max()) + 1 # Number of clusters
-        nn = len(self.adata) # Number of observations
-        self.logger.info('data firms={} workers={} clusters={} observations={}'.format(self.nf, self.nw, self.nc, nn))
+        self.logger.info(f'data firms={self.nf} workers={self.nw} clusters={self.nc} observations={len(self.adata)}')
 
-        nm = len(np.unique(self.adata[self.adata['m'] == 1]['i'])) # Number of movers
+        nm = self.adata.loc[self.adata['m'] == 1, 'i'].nunique() # Number of movers
         self.ns = self.nw - nm # Number of stayers
-        self.logger.info('data movers={} stayers={}'.format(nm, self.ns))
+        self.logger.info(f'data movers={nm} stayers={self.ns}')
 
         self.res['n_firms'] = self.nf
         self.res['n_workers'] = self.nw
@@ -274,9 +305,9 @@ class CREEstimator:
         if self.wo_btw:
             EEm = EEm * 0.0
         jdata = pd.merge(jdata, EEm, on=('g1', 'g2'))
-        pb['EEm'] = pd_to_np(EEm.reset_index(), 'g1', 'g2', 'mx', self.nc, self.nc)
+        pb['EEm'] = _pd_to_np(EEm.reset_index(), 'g1', 'g2', 'mx', self.nc, self.nc)
         #pb['EEm'] = np.array(EEm.values).reshape(self.nc, self.nc)
-        #print(pd_to_np(EEm.reset_index(), 'g1', 'g2', 'mx', self.nc, self.nc) - np.array(EEm.values).reshape(self.nc, self.nc))
+        #print(_pd_to_np(EEm.reset_index(), 'g1', 'g2', 'mx', self.nc, self.nc) - np.array(EEm.values).reshape(self.nc, self.nc))
 
         sdata['psi1_tmp'] = pb['Afill'][sdata['g1']]
         Em = sdata.assign(mx = lambda df: df['y1'] - df['psi1_tmp']).groupby(['g1'])['mx'].agg('mean')
@@ -296,7 +327,7 @@ class CREEstimator:
 
         return sdata, jdata
 
-    def __estimate_within_cluster(self, sdata, jdata):
+    def _estimate_within_cluster(self, sdata, jdata):
         '''
         @ FIXME add description of function
 
@@ -344,7 +375,7 @@ class CREEstimator:
         dm.eval('y2m2j_lo = (nm2j * y2m2j - nm2c * y2m2c) / (nm2j - nm2c)', inplace=True)
 
         # Compute the moments involving stayers
-        res['y1s_y1s'] = ds.query('nsj > 1').pipe(pipe_qcov, 'y1n', 'y1s_lo')
+        res['y1s_y1s'] = ds.query('nsj > 1').pipe(_pipe_qcov, 'y1n', 'y1s_lo')
         res['y1s_y1s_count'] = ds.query('nsj > 1').shape[0]
         res['y1s_var'] = ds['y1n'].var()
         res['y1s_var_count'] = ds.shape[0]
@@ -354,33 +385,33 @@ class CREEstimator:
         res['y2m_var_count'] = dm.shape[0]
 
         # Compute the moments involving movers leaving the firm
-        res['y1s_y1m1'] = ds.query('nm1j > 0').pipe(pipe_qcov, 'y1n', 'y1m1j')
+        res['y1s_y1m1'] = ds.query('nm1j > 0').pipe(_pipe_qcov, 'y1n', 'y1m1j')
         res['y1s_y1m1_count'] = ds.query('nm1j > 0').shape[0]
-        res['y1s_y2m1'] = ds.query('nm1j > 0').pipe(pipe_qcov, 'y1n', 'y2m1j')
+        res['y1s_y2m1'] = ds.query('nm1j > 0').pipe(_pipe_qcov, 'y1n', 'y2m1j')
         res['y1s_y2m1_count'] = ds.query('nm1j > 0').shape[0]
-        res['y1m1_y1m1'] = dm.query('nm1j > nm1c').pipe(pipe_qcov, 'y1n', 'y1m1j_lo')
+        res['y1m1_y1m1'] = dm.query('nm1j > nm1c').pipe(_pipe_qcov, 'y1n', 'y1m1j_lo')
         res['y1m1_y1m1_count'] = dm.query('nm1j > nm1c').shape[0]
-        res['y2m1_y1m1'] = dm.query('nm1j > nm1c').pipe(pipe_qcov, 'y2n', 'y1m1j_lo')
+        res['y2m1_y1m1'] = dm.query('nm1j > nm1c').pipe(_pipe_qcov, 'y2n', 'y1m1j_lo')
         res['y2m1_y1m1_count'] = dm.query('nm1j > nm1c').shape[0]
-        res['y2m1_y2m1'] = dm.query('nm1j > nm1c').pipe(pipe_qcov, 'y2n', 'y2m1j_lo')
+        res['y2m1_y2m1'] = dm.query('nm1j > nm1c').pipe(_pipe_qcov, 'y2n', 'y2m1j_lo')
         res['y2m1_y2m1_count'] = dm.query('nm1j > nm1c').shape[0]
 
         # Compute the moments involving movers arriving at the firm
-        res['y1s_y1m2'] = ds.query('nm2j > 0').pipe(pipe_qcov, 'y1n', 'y1m2j')
+        res['y1s_y1m2'] = ds.query('nm2j > 0').pipe(_pipe_qcov, 'y1n', 'y1m2j')
         res['y1s_y1m2_count'] = ds.query('nm2j > 0').shape[0]
-        res['y1s_y2m2'] = ds.query('nm2j > 0').pipe(pipe_qcov, 'y1n', 'y2m2j')
+        res['y1s_y2m2'] = ds.query('nm2j > 0').pipe(_pipe_qcov, 'y1n', 'y2m2j')
         res['y1s_y2m2_count'] = ds.query('nm2j > 0').shape[0]
-        res['y1m2_y1m2'] = dm.query('nm2j > nm2c').pipe(pipe_qcov, 'y1n', 'y1m2j_lo')
+        res['y1m2_y1m2'] = dm.query('nm2j > nm2c').pipe(_pipe_qcov, 'y1n', 'y1m2j_lo')
         res['y1m2_y1m2_count'] = dm.query('nm2j > nm2c').shape[0]
-        res['y2m2_y1m2'] = dm.query('nm2j > nm2c').pipe(pipe_qcov, 'y2n', 'y1m2j_lo')
+        res['y2m2_y1m2'] = dm.query('nm2j > nm2c').pipe(_pipe_qcov, 'y2n', 'y1m2j_lo')
         res['y2m2_y1m2_count'] = dm.query('nm2j > nm2c').shape[0]
-        res['y2m2_y2m2'] = dm.query('nm2j > nm2c').pipe(pipe_qcov, 'y2n', 'y2m2j_lo')
+        res['y2m2_y2m2'] = dm.query('nm2j > nm2c').pipe(_pipe_qcov, 'y2n', 'y2m2j_lo')
         res['y2m2_y2m2_count'] = dm.query('nm2j > nm2c').shape[0]
 
         # Total variance of wages in differences for movers
         res['dym_dym'] = dm.query('g1 != g2').eval('y2n-y1n').var()
         res['dym_dym_count'] = dm.query('g1 != g2').shape[0]
-        res['y1m_y2m'] = dm.query('g1 != g2').pipe(pipe_qcov, 'y1n', 'y2n')
+        res['y1m_y2m'] = dm.query('g1 != g2').pipe(_pipe_qcov, 'y1n', 'y2n')
         res['y1m_y2m_count'] = dm.query('g1 != g2').shape[0]
 
         self.moments_within = res
@@ -456,7 +487,7 @@ class CREEstimator:
             Yq (Pandas Series): income for movers
         '''
         self.res['var_y'] = Yq.var()
-        self.logger.info('total variance: {:0.4f}'.format(self.res['var_y']))
+        self.logger.info(f"total variance: {self.res['var_y']:0.4f}")
 
         # Compute the between terms
         cov_mat_between = cdata.cov()
@@ -470,8 +501,8 @@ class CREEstimator:
         self.res['tot_cov'] = self.res['cov_bw'] + self.res['cov_wt']
         self.res['var_y'] = np.var(Yq)
 
-        self.logger.info('[cre] VAR bw={:4f} wt={:4f} tot={:4f}'.format(self.res['var_bw'], self.res['var_wt'], self.res['var_bw'] + self.res['var_wt']))
-        self.logger.info('[cre] COV bw={:4f} wt={:4f} tot={:4f}'.format(self.res['cov_bw'], self.res['cov_wt'], self.res['cov_bw'] + self.res['cov_wt']))
+        self.logger.info(f"[cre] VAR bw={self.res['var_bw']:4f} wt={self.res['var_wt']:4f} tot={self.res['var_bw'] + self.res['var_wt']:4f}")
+        self.logger.info(f"[cre] COV bw={self.res['cov_bw']:4f} wt={self.res['cov_wt']:4f} tot={self.res['cov_bw'] + self.res['cov_wt']:4f}")
 
         # Create summary variable
         self.summary['var_y'] = self.res['var_y']
@@ -488,7 +519,7 @@ class CREEstimator:
             jdata (Pandas DataFrame): movers
         '''
         jdata_f = pd.concat([jdata[['j1', 'g1']], jdata[['j2', 'g2']].rename(columns={'j2': 'j1', 'g2': 'g1'})]).drop_duplicates()
-        Jf = csc_matrix((np.ones(len(jdata_f)), (jdata_f['j1'], jdata_f['g1'])), shape=(len(jdata_f), self.nc))
+        Jf = csc_matrix((np.ones(len(jdata_f)), (jdata_f['j1'], jdata_f['g1'])), shape=(self.nf, self.nc)) # Need self.nf because recall there can be firms in j2 that aren't in j1
         self.Mud = Jf * self.between_params['Afill']
 
     def __prep_posterior_var(self, jdata, cdata):
@@ -519,10 +550,16 @@ class CREEstimator:
         M = 1 / self.within_params['var_psi'] * eye(self.nf) + 1 / self.within_params['var_eps'] * Jd.transpose() * Jd
         self.ml = pyamg.ruge_stuben_solver(M)
 
-    def __compute_posterior_var(self):
+    def __compute_posterior_var(self, rng=None):
         '''
         Compute the posterior variance of the CRE model.
+
+        Arguments:
+            rng (np.random.Generator or None): NumPy random number generator; None is equivalent to np.random.default_rng(None)
         '''
+        if rng is None:
+            rng = np.random.default_rng(None)
+
         # We first compute the direct term
         v1 = 1 / self.within_params['var_psi'] * self.Mud
         v1 += 1 / self.within_params['var_eps'] * self.Jd.transpose() * self.Yd
@@ -533,8 +570,10 @@ class CREEstimator:
 
         # Next we look at the trace term
         tr_var_pos_all = np.zeros(self.ndraw_trace)
-        for r in trange(self.ndraw_trace):
-            Zpsi = 2 * np.random.binomial(1, 0.5, self.nf) - 1
+        pbar = trange(self.ndraw_trace)
+        pbar.set_description('cre')
+        for r in pbar:
+            Zpsi = 2 * rng.binomial(1, 0.5, self.nf) - 1
 
             R1 = self.Jq * Zpsi
             R2 = self.Jq * self.ml.solve(Zpsi)
@@ -543,17 +582,21 @@ class CREEstimator:
         t2 = np.mean(tr_var_pos_all)
         self.posterior_var = t1 + t2
 
-        self.logger.info('[cre] posterior variance of psi = {:4f}'.format(self.posterior_var))
+        self.logger.info(f'[cre] posterior variance of psi = {self.posterior_var:4f}')
 
     def __save_res(self):
         '''
         Save results as json.
         '''
-        # Convert results into strings to prevent JSON errors
-        for key, val in self.res.items():
-            self.res[key] = str(val)
+        outputfile = self.params['outputfile']
+        if outputfile is not None:
+            # Convert results into strings to prevent JSON errors
+            for key, val in self.res.items():
+                self.res[key] = str(val)
 
-        with open(self.params['out'], 'w') as outfile:
-            json.dump(self.res, outfile)
+            with open(outputfile, 'w') as outfile:
+                json.dump(self.res, outfile)
 
-        self.logger.info('saved results to {}'.format(self.params['out']))
+            self.logger.info(f'saved results to {outputfile}')
+        else:
+            self.logger.info('outputfile=None, so results not saved')
